from kooplearn.datasets.deterministic import DuffingOscillator, Lorenz63
from kooplearn.datasets.stochastic import (
    LangevinTripleWell1D,
    LinearModel,
    LogisticMap,
    Mock,
    MullerBrownPotential,
)
